###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order iron and manganese removal model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    Block, check_optimal_termination, ConcreteModel, Constraint, value, Var)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import IronManganeseRemovalZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()

class TestIronManganeseRemovalZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["iron", "manganese", "foo"]})

        m.fs.unit = IronManganeseRemovalZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "iron"].fix(250)
        m.fs.unit.inlet.flow_mass_comp[0, "manganese"].fix(250)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert isinstance(model.fs.unit.electricity_constraint,
                          Constraint)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_intensity_parameter, Var)
        assert isinstance(model.fs.unit.air_water_ratio, Var)
        assert isinstance(model.fs.unit.flow_basis, Var)
        assert isinstance(model.fs.unit.air_flow_rate, Var)
        assert isinstance(model.fs.unit.filter_surf_area, Var)
        assert isinstance(model.fs.unit.num_filter_units, Var)


    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("iron_and_manganese_removal")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data[
                    "default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.air_water_ratio[0].fixed
        assert model.fs.unit.air_water_ratio[0].value == data[
            "air_water_ratio"]["value"]

        assert model.fs.unit.flow_basis[0].fixed
        assert model.fs.unit.flow_basis[0].value == data[
            "flow_basis"]["value"]

        assert model.fs.unit.electricity_intensity_parameter.fixed
        assert model.fs.unit.electricity_intensity_parameter.value == data[
            "electricity_intensity_parameter"]["value"]

        assert model.fs.unit.filter_surf_area.fixed
        assert model.fs.unit.filter_surf_area.value == data[
            "filter_surf_area"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(10.501, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(23.8073, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["iron"]))
        assert (pytest.approx(23.8073, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["manganese"]))
        assert (pytest.approx(0.095229, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["foo"]))
        assert (pytest.approx(10.050, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(2.48756, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["iron"]))
        assert (pytest.approx(2.48756, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["manganese"]))
        assert (pytest.approx(0.099502, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0.45100, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(498.89, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["iron"]))
        assert (pytest.approx(498.89, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["manganese"]))
        assert (pytest.approx(2.2173e-08, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]))
        assert (pytest.approx(4166.50264, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                                : Value   : Fixed : Bounds
    Electricity intensity per Inlet Flowrate  (kWh/m3) : 0.11021 : False : (None, None)
                                Power Consumption (kW) :  4166.5 : False : (0, None)
                                  Solute Removal [foo] :  0.0000 :  True : (0, None)
                                 Solute Removal [iron] : 0.90000 :  True : (0, None)
                            Solute Removal [manganese] : 0.90000 :  True : (0, None)
                                        Water Recovery : 0.99990 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                   Inlet   Treated  Byproduct
    Volumetric Flowrate            10.501   10.050    0.45100
    Mass Concentration H2O         952.29   994.93     2.2173
    Mass Concentration iron        23.807   2.4876     498.89
    Mass Concentration manganese   23.807   2.4876     498.89
    Mass Concentration foo       0.095229 0.099502 2.2173e-08
====================================================================================
"""
        assert output in stream.getvalue()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["iron", "manganese", "foo"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = IronManganeseRemovalZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "iron"].fix(250)
    m.fs.unit1.inlet.flow_mass_comp[0, "manganese"].fix(250)
    m.fs.unit1.inlet.flow_mass_comp[0, "foo"].fix(1)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.iron_and_manganese_removal, Block)
    assert isinstance(m.fs.costing.iron_and_manganese_removal.capital_blower_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.iron_and_manganese_removal.capital_backwash_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.iron_and_manganese_removal.capital_backwash_b_parameter,
                      Var)
    assert isinstance(m.fs.costing.iron_and_manganese_removal.capital_filter_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.iron_and_manganese_removal.capital_filter_b_parameter,
                      Var)
    assert isinstance(m.fs.costing.iron_and_manganese_removal.flow_exponent,
                      Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order filter press model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    Block, check_optimal_termination, ConcreteModel, Constraint, value, Var)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import FilterPressZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestFilterPressZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(default={
                "solute_list": ["tss"]})


        m.fs.unit = FilterPressZO(default={"property_package": m.fs.params,
                                    "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(23)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert model.fs.unit._tech_type == "filter_press"
        assert isinstance(model.fs.unit.hours_per_day_operation, Var)
        assert isinstance(model.fs.unit.cycle_time, Var)
        assert isinstance(model.fs.unit.electricity_a_parameter, Var)
        assert isinstance(model.fs.unit.electricity_b_parameter, Var)
        assert isinstance(model.fs.unit.filter_press_capacity, Var)
        assert isinstance(model.fs.unit.fp_electricity, Constraint)
        assert isinstance(model.fs.unit.fp_capacity, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("filter_press")
        model.fs.unit.load_parameters_from_database()
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 0.0001

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j not in data["removal_frac_mass_solute"]:
                assert v.value == data[
                    "default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.hours_per_day_operation[0].fixed
        assert model.fs.unit.hours_per_day_operation[0].value == data[
            "hours_per_day_operation"]["value"]
        assert model.fs.unit.cycle_time[0].fixed
        assert model.fs.unit.cycle_time[0].value == data[
            "cycle_time"]["value"]
        assert model.fs.unit.electricity_a_parameter[0].fixed
        assert model.fs.unit.electricity_a_parameter[0].value == data[
            "electricity_a_parameter"]["value"]
        assert model.fs.unit.electricity_b_parameter[0].fixed
        assert model.fs.unit.electricity_b_parameter[0].value == data[
            "electricity_b_parameter"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.024, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(41.66666, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp[
                "H2O"]))
        assert (pytest.approx(958.33333, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp[
                "tss"]))

        assert (pytest.approx(0.000460, rel=1e-3) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.217344, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "H2O"]))
        assert (pytest.approx(999.782655, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "tss"]))

        assert (pytest.approx(0.0235399, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(42.476815, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp[
                "H2O"]))
        assert (pytest.approx(957.523184, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp[
                "tss"]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))
    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                         : Value      : Fixed : Bounds
    Filter Press Capacity (ft3) :     9153.6 : False : (None, None)
        Filter Press Power (kW) :     156.61 : False : (0, None)
           Solute Removal [tss] :    0.98000 :  True : (0, None)
                 Water Recovery : 0.00010000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                             Inlet    Treated   Byproduct
    Volumetric Flowrate    0.024000 0.00046010  0.023540 
    Mass Concentration H2O   41.667    0.21734    42.477 
    Mass Concentration tss   958.33     999.78    957.52 
====================================================================================
"""

        assert output in stream.getvalue()

def test_costing():

    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.params = WaterParameterBlock(default={
            "solute_list": ["tss"]})

    m.fs.unit = FilterPressZO(default={"property_package": m.fs.params,
                                "database": m.db})

    m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1)
    m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(23)

    m.fs.costing = ZeroOrderCosting()
    m.fs.unit.load_parameters_from_database()

    m.fs.unit.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.filter_press, Block)
    assert isinstance(m.fs.costing.filter_press.capital_a_parameter, Var)
    assert isinstance(m.fs.costing.filter_press.capital_b_parameter, Var)

    assert isinstance(m.fs.unit.costing.capital_cost, Var)
    assert isinstance(m.fs.unit.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit) == 0

    assert m.fs.unit.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order cartridge filtration model
"""
import pytest

from io import StringIO
from pyomo.environ import (
    Block, ConcreteModel, Constraint, value, Var, assert_optimal_termination)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import CartridgeFiltrationZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()

class TestCartridgeFiltrationZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["nonvolatile_toc", "tss"]})

        m.fs.unit = CartridgeFiltrationZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == 'cartridge_filtration'
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("cartridge_filtration")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(1.2e-2, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(83.3333, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(83.3333, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.010899, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(73.4012, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(9.17515, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(1.101e-3, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(181.653, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(817.439, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.00864, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                              : Value      : Fixed : Bounds
                  Electricity Demand :  0.0086400 : False : (0, None)
               Electricity Intensity : 0.00020000 :  True : (None, None)
    Solute Removal [nonvolatile_toc] :    0.20000 :  True : (0, None)
                Solute Removal [tss] :    0.90000 :  True : (0, None)
                      Water Recovery :    0.99990 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                         Inlet   Treated  Byproduct
    Volumetric Flowrate                0.012000 0.010899 0.0011010 
    Mass Concentration H2O               833.33   917.42   0.90827 
    Mass Concentration nonvolatile_toc   83.333   73.401    181.65 
    Mass Concentration tss               83.333   9.1752    817.44 
====================================================================================
"""

        assert output in stream.getvalue()

class TestCartridgeFiltrationZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["nonvolatile_toc", "tss", "foo"]})

        m.fs.unit = CartridgeFiltrationZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == 'cartridge_filtration'
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("cartridge_filtration")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(1.3e-2, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(76.923, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(76.923, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["tss"]))
        assert (pytest.approx(76.923, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0.011899, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(67.2325, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(8.40407, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(84.0407, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(1.101e-3, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(181.653, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(817.439, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(7.26612e-7, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0.00936, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                              : Value      : Fixed : Bounds
                  Electricity Demand :  0.0093600 : False : (0, None)
               Electricity Intensity : 0.00020000 :  True : (None, None)
                Solute Removal [foo] :     0.0000 :  True : (0, None)
    Solute Removal [nonvolatile_toc] :    0.20000 :  True : (0, None)
                Solute Removal [tss] :    0.90000 :  True : (0, None)
                      Water Recovery :    0.99990 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                         Inlet   Treated  Byproduct
    Volumetric Flowrate                0.013000 0.011899  0.0011010
    Mass Concentration H2O               769.23   840.32    0.90827
    Mass Concentration nonvolatile_toc   76.923   67.233     181.65
    Mass Concentration tss               76.923   8.4041     817.44
    Mass Concentration foo               76.923   84.041 7.2661e-07
====================================================================================
"""

        assert output in stream.getvalue()

def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = CartridgeFiltrationZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.cartridge_filtration, Block)
    assert isinstance(m.fs.costing.cartridge_filtration.capital_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.cartridge_filtration.capital_b_parameter,
                      Var)
    assert isinstance(m.fs.costing.cartridge_filtration.reference_state, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

from enum import Enum

import pyomo.environ as pyo

from pyomo.util.calc_var_value import calculate_variable_from_constraint

from idaes.core.util.exceptions import BurntToast, ConfigurationError
from idaes.core.util.misc import StrEnum
from idaes.core import declare_process_block_class
from idaes.generic_models.costing.costing_base import (
    FlowsheetCostingBlockData, register_idaes_currency_units)

from idaes.generic_models.unit_models import Mixer

from watertap.unit_models import (
        ReverseOsmosis0D,
        ReverseOsmosis1D,
        NanoFiltration0D,
        NanofiltrationZO,
        PressureExchanger,
        Pump,
        )


class ROType(StrEnum):
    standard = "standard"
    high_pressure = "high_pressure"


class PumpType(StrEnum):
    low_pressure = "low_pressure"
    high_pressure = "high_pressure"
    pressure_exchanger = "pressure_exchanger"
    energy_recovery_device = "energy_recovery_device"


class MixerType(StrEnum):
    default = "default"
    NaOCl = "NaOCl"
    CaOH2 = "CaOH2"


@declare_process_block_class("WaterTAPCosting")
class WaterTAPCostingData(FlowsheetCostingBlockData):

    def build_global_params(self):

        # Register currency and conversion rates based on CE Index
        register_idaes_currency_units()

        # Set the base year for all costs
        self.base_currency = pyo.units.USD_2018 
        # Set a base period for all operating costs
        self.base_period = pyo.units.year

        # Build flowsheet level costing components
        # This is package specific
        self.load_factor = pyo.Var(
                initialize=0.9,
                doc='Load factor [fraction of uptime]',
                units=pyo.units.dimensionless)
        self.factor_total_investment = pyo.Var(
                initialize=2,
                doc='Total investment factor [investment cost/equipment cost]',
                units=pyo.units.dimensionless)
        self.factor_maintenance_labor_chemical = pyo.Var(
                initialize=0.03,
                doc='Maintenance-labor-chemical factor [fraction of investment cost/year]',
                units=pyo.units.year**-1)
        self.factor_capital_annualization = pyo.Var(
                initialize=0.1,
                doc='Capital annualization factor [fraction of investment cost/year]',
                units=pyo.units.year**-1)
        self.factor_membrane_replacement = pyo.Var(
                initialize=0.2,
                doc='Membrane replacement factor [fraction of membrane replaced/year]',
                units=pyo.units.year**-1)
        self.reverse_osmosis_membrane_cost = pyo.Var(
                initialize=30,
                doc='Membrane cost',
                units=self.base_currency/(pyo.units.meter**2))
        self.reverse_osmosis_high_pressure_membrane_cost = pyo.Var(
                initialize=75,
                doc='Membrane cost',
                units=self.base_currency/(pyo.units.meter**2))
        self.nanofiltration_membrane_cost = pyo.Var(
                initialize=15,
                doc='Membrane cost',
                units=self.base_currency/(pyo.units.meter**2))
        self.high_pressure_pump_cost = pyo.Var(
                initialize=53 / 1e5 * 3600,
                doc='High pressure pump cost',
                units=self.base_currency/pyo.units.watt)
        self.low_pressure_pump_cost = pyo.Var(
                initialize=889,
                doc='Low pressure pump cost',
                units=self.base_currency/(pyo.units.liter/pyo.units.second))
        self.pump_pressure_exchanger_cost = pyo.Var(
                initialize=535,
                doc='Pressure exchanger cost',
                units=self.base_currency/(pyo.units.meter**3/pyo.units.hours))
        self.pressure_exchanger_cost = pyo.Var(
                initialize=535,
                doc='Pressure exchanger cost',
                units=self.base_currency/(pyo.units.meter**3/pyo.units.hours))
        self.energy_recovery_device_linear_coefficient = pyo.Var(
                initialize=3134.7,
                doc='Energy recovery device linear coefficient',
                units=self.base_currency)
        self.energy_recovery_device_exponent = pyo.Var(
                initialize=0.58,
                doc='Energy recovery device exponent',
                units=pyo.units.dimensionless)
        self.mixer_unit_cost = pyo.Var(
                initialize=361,
                doc='Mixer cost',
                units=self.base_currency/(pyo.units.liters/pyo.units.second))
        self.naocl_mixer_unit_cost = pyo.Var(
                initialize=5.08,
                doc='NaOCl mixer cost',
                units=self.base_currency/(pyo.units.m**3/pyo.units.day))
        self.caoh2_mixer_unit_cost = pyo.Var(
                initialize=792.8*2.20462,
                doc='Ca(OH)2 mixer cost',
                units=self.base_currency/(pyo.units.kg/pyo.units.day))

        self.electricity_base_cost = pyo.Param(
                mutable=True,
                initialize=0.07,
                doc='Electricity cost',
                units=self.base_currency/pyo.units.kWh)
        self.naocl_cost = pyo.Param(
                initialize=0.23,
                doc='NaOCl cost',
                units=self.base_currency/pyo.units.kg)
        self.naocl_purity = pyo.Param(
                mutable=True,
                initialize=0.15,
                doc='NaOCl purity',
                units=pyo.units.dimensionless)
        self.caoh2_cost = pyo.Param(
                mutable=True,
                initialize=0.12,
                doc='CaOH2 cost',
                units=self.base_currency/pyo.units.kg)
        self.caoh2_purity = pyo.Param(
                mutable=True,
                initialize=1,
                doc='CaOH2 purity',
                units=pyo.units.dimensionless)

        # fix the parameters
        for var in self.component_objects(pyo.Var):
            var.fix()

        # Define standard material flows and costs
        self.defined_flows["electricity"] = self.electricity_base_cost
        self.defined_flows["NaOCl"] = 74.44e-3 * (pyo.units.kg/pyo.units.mol) * (self.naocl_cost/self.naocl_purity)
        self.defined_flows["CaOH2"] = 74.093e-3 * (pyo.units.kg/pyo.units.mol) * (self.caoh2_cost/self.caoh2_purity)


    def build_process_costs(self):
        self.total_capital_cost = pyo.Expression(
                expr = self.aggregate_capital_cost,
                doc='Total capital cost')
        self.total_investment_cost = pyo.Var(
                initialize=1e3,
                domain=pyo.NonNegativeReals,
                doc='Total investment cost',
                units=self.base_currency)
        self.maintenance_labor_chemical_operating_cost = pyo.Var(
                initialize=1e3,
                domain=pyo.NonNegativeReals,
                doc='Maintenance-labor-chemical operating cost',
                units=self.base_currency/self.base_period)
        self.total_operating_cost= pyo.Var(
                initialize=1e3,
                domain=pyo.NonNegativeReals,
                doc='Total operating cost',
                units=self.base_currency/self.base_period)

        self.total_investment_cost_constraint = pyo.Constraint(expr = \
                self.total_investment_cost == self.factor_total_investment * self.total_capital_cost)
        self.maintenance_labor_chemical_operating_cost_constraint = pyo.Constraint(expr = \
                self.maintenance_labor_chemical_operating_cost == self.factor_maintenance_labor_chemical * self.total_investment_cost)

        self.total_operating_cost_constraint = pyo.Constraint(expr = \
                self.total_operating_cost == self.maintenance_labor_chemical_operating_cost \
                + self.aggregate_fixed_operating_cost \
                + self.aggregate_variable_operating_cost \
                + sum(self.aggregate_flow_costs.values())*self.load_factor )

    def initialize_build(self):
        calculate_variable_from_constraint(self.total_investment_cost, self.total_investment_cost_constraint)
        calculate_variable_from_constraint(self.maintenance_labor_chemical_operating_cost,
                self.maintenance_labor_chemical_operating_cost_constraint)
        calculate_variable_from_constraint(self.total_operating_cost,
                self.total_operating_cost_constraint)

        if hasattr(self, "LCOW"):
            calculate_variable_from_constraint(self.LCOW, self.LCOW_constraint)

    def add_LCOW(self, flow_rate):
        """
        Add Levelized Cost of Water (LCOW) to costing block.
        Args:
            flow_rate - flow rate of water (volumetric) to be used in
                        calculating LCOW
        """

        self.annual_water_production = pyo.Expression( expr=
                 (pyo.units.convert(
                     flow_rate,
                     to_units=pyo.units.m**3/self.base_period) *
                  self.load_factor))

        self.LCOW = pyo.Var(
            doc='Levelized Cost of Water',
            units=self.base_currency/pyo.units.m**3)

        self.LCOW_constraint = pyo.Constraint(expr = self.LCOW == 
            (self.total_investment_cost*self.factor_capital_annualization +
                  self.total_operating_cost) / self.annual_water_production)

    def add_specific_energy_consumption(self, flow_rate):
        """
        Add specific energy consumption (kWh/m**3) to costing block.
        Args:
            flow_rate - flow rate of water (volumetric) to be used in
                        calculating specific energy consumption
        """

        self.specific_energy_consumption = pyo.Expression(expr = 
                self.aggregate_flow_electricity / pyo.units.convert(flow_rate, to_units=pyo.units.m**3/pyo.units.hr))

    # Define costing methods supported by package
    @staticmethod
    def cost_nanofiltration(blk):
        """
        Nanofiltration costing method

        TODO: describe equations
        """
        cost_membrane(blk, blk.costing_package.nanofiltration_membrane_cost, blk.costing_package.factor_membrane_replacement)

    @staticmethod
    def cost_reverse_osmosis(blk, ro_type=ROType.standard):
        """
        Reverse osmosis costing method

        TODO: describe equations

        Args:
            ro_type - ROType Enum indicating reverse osmosis type,
                      default = ROType.standard
        """
        # Validate arguments
        if ro_type not in ROType:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for ro_type:"
                f" {ro_type}. Argument must be a member of the ROType Enum.")

        if ro_type == ROType.standard:
            membrane_cost = blk.costing_package.reverse_osmosis_membrane_cost
        elif ro_type == ROType.high_pressure:
            membrane_cost = blk.costing_package.reverse_osmosis_high_pressure_membrane_cost
        else:
            raise BurntToast(f"Unrecognized ro_type: {ro_type}")
        cost_membrane(blk, membrane_cost, blk.costing_package.factor_membrane_replacement)

    @staticmethod
    def cost_pump(blk, pump_type=PumpType.high_pressure):
        """
        Pump costing method

        TODO: describe equations

        Args:
            pump_type - PumpType Enum indicating pump type,
                        default = pumptype.high_pressure
        """
        if pump_type not in PumpType:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for pump_type:"
                f" {pump_type}. Argument must be a member of the PumpType Enum.")

        if pump_type == PumpType.high_pressure:
            WaterTAPCostingData.cost_high_pressure_pump(blk)
        elif pump_type == PumpType.low_pressure:
            WaterTAPCostingData.cost_low_pressure_pump(blk)
        elif pump_type == PumpType.pressure_exchanger:
            WaterTAPCostingData.cost_pressure_exchanger_pump(blk)
        elif pump_type == PumpType.energy_recovery_device:
            WaterTAPCostingData.cost_energy_recovery_device_pump(blk)
        else:
            raise BurntToast(f"Unrecognized pump_type: {pump_type}")

    @staticmethod
    def cost_high_pressure_pump(blk):
        """
        High pressure pump costing method

        TODO: describe equations
        """
        _make_capital_cost_var(blk)
        blk.capital_cost_constraint = pyo.Constraint(expr = \
                blk.capital_cost == blk.costing_package.high_pressure_pump_cost * pyo.units.convert(blk.unit_model.work_mechanical[0], pyo.units.W))

    @staticmethod
    def cost_low_pressure_pump(blk):
        """
        Low pressure pump costing method

        TODO: describe equations
        """
        cost_by_flow_volume(blk, blk.costing_package.low_pressure_pump_cost,
                pyo.units.convert(blk.unit_model.control_volume.properties_in[0].flow_vol, (pyo.units.m**3/pyo.units.s)))

    @staticmethod
    def cost_pressure_exchanger_pump(blk):
        """
        Pump pressure exchanger costing method

        TODO: describe equations
        """
        cost_by_flow_volume(blk, blk.costing_package.pump_pressure_exchanger_cost,
                pyo.units.convert(blk.unit_model.control_volume.properties_in[0].flow_vol, (pyo.units.meter**3/pyo.units.hours)))

    @staticmethod
    def cost_energy_recovery_device_pump(blk):
        """
        Pump energy recovery device costing method

        TODO: describe equations
        """
        _make_capital_cost_var(blk)
        unit_cv_in = blk.unit_model.control_volume.properties_in[0]
        blk.capital_cost_constraint = pyo.Constraint(expr = \
                  blk.capital_cost == blk.costing_package.energy_recovery_device_linear_coefficient
                 * (pyo.units.convert((sum(unit_cv_in.flow_mass_phase_comp['Liq', j]
                        for j in blk.unit_model.config.property_package.component_list)
                 / unit_cv_in.dens_mass_phase['Liq']), pyo.units.m**3/pyo.units.hour) / (pyo.units.m**3/pyo.units.hour)) ** blk.costing_package.energy_recovery_device_exponent)

    @staticmethod
    def cost_pressure_exchanger(blk):
        """
        Pressure exchanger costing method

        TODO: describe equations
        """
        cost_by_flow_volume(blk, blk.costing_package.pressure_exchanger_cost,
                pyo.units.convert(blk.unit_model.low_pressure_side.properties_in[0].flow_vol, (pyo.units.meter**3/pyo.units.hours)))

    @staticmethod
    def cost_mixer(blk, mixer_type=MixerType.default):
        """
        Mixer costing method

        TODO: describe equations

        args:
            mixer_type - MixerType Enum indicating mixer type,
                        default = MixerType.default
        """
        if mixer_type not in MixerType:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for mixer_type:"
                f" {mixer_type}. Argument must be a member of the MixerType Enum.")

        if mixer_type==MixerType.default:
            WaterTAPCostingData.cost_default_mixer(blk)
        elif mixer_type==MixerType.NaOCl:
            WaterTAPCostingData.cost_naocl_mixer(blk)
        elif mixer_type==MixerType.CaOH2:
            WaterTAPCostingData.cost_caoh2_mixer(blk)
        else:
            raise BurntToast(f"Unrecognized mixer_type: {mixer_type}")

    @staticmethod
    def cost_default_mixer(blk):
        """
        Default mixer costing method

        TODO: describe equations
        """
        cost_by_flow_volume(blk, blk.costing_package.mixer_unit_cost,
                pyo.units.convert(blk.unit_model.mixed_state[0].flow_vol, pyo.units.liter/pyo.units.second))

    @staticmethod
    def cost_naocl_mixer(blk):
        """
        NaOCl mixer costing method

        TODO: describe equations
        """
        cost_by_flow_volume(blk, blk.costing_package.naocl_mixer_unit_cost,
                pyo.units.convert(blk.unit_model.inlet_stream_state[0].flow_vol, pyo.units.m**3/pyo.units.day))

    @staticmethod
    def cost_caoh2_mixer(blk):
        """
        CaOH2 mixer costing method

        TODO: describe equations
        """
        stream = blk.unit_model.lime_stream
        blk.lime_kg_per_day = pyo.Expression(expr=\
                pyo.units.convert(74.093e-3 * (pyo.units.kg/pyo.units.mol) * stream.flow_mol[0]*stream.mole_frac_comp[0, "Ca(OH)2"],
                    pyo.units.kg/pyo.units.day))
        cost_by_flow_volume(blk, blk.costing_package.caoh2_mixer_unit_cost/blk.costing_package.factor_total_investment,
                blk.lime_kg_per_day)


# Define default mapping of costing methods to unit models
WaterTAPCostingData.unit_mapping = {
        Mixer: WaterTAPCostingData.cost_mixer,
        Pump: WaterTAPCostingData.cost_pump,
        PressureExchanger: WaterTAPCostingData.cost_pressure_exchanger,
        ReverseOsmosis0D: WaterTAPCostingData.cost_reverse_osmosis,
        ReverseOsmosis1D: WaterTAPCostingData.cost_reverse_osmosis,
        NanoFiltration0D: WaterTAPCostingData.cost_nanofiltration,
        NanofiltrationZO: WaterTAPCostingData.cost_nanofiltration,
        }

def _make_capital_cost_var(blk):
    blk.capital_cost = pyo.Var(initialize=1e5,
                           domain=pyo.NonNegativeReals,
                           units=blk.costing_package.base_currency,
                           doc="Unit capital cost")

def _make_fixed_operating_cost_var(blk):
    blk.fixed_operating_cost = pyo.Var(initialize=1e5,
                                   domain=pyo.NonNegativeReals,
                                   units=blk.costing_package.base_currency/blk.costing_package.base_period,
                                   doc="Unit fixed operating cost")

def cost_membrane(blk, membrane_cost, factor_membrane_replacement):
    """
    Generic function for costing a membrane. Assumes the unit_model
    has an `area` variable or parameter.

    Args:
        membrane_cost - The cost of the membrane in currency per area
        factor_membrane_replacement - Membrane replacement factor
                                      [fraction of membrane replaced/year]

    """
    _make_capital_cost_var(blk)
    _make_fixed_operating_cost_var(blk)

    blk.membrane_cost = pyo.Expression(expr=membrane_cost)
    blk.factor_membrane_replacement = pyo.Expression(expr=factor_membrane_replacement)

    blk.capital_cost_constraint = pyo.Constraint(expr = \
            blk.capital_cost == blk.membrane_cost * pyo.units.convert(blk.unit_model.area, pyo.units.m**2))
    blk.fixed_operating_cost_constraint = pyo.Constraint(expr = \
            blk.fixed_operating_cost == blk.factor_membrane_replacement * blk.membrane_cost * pyo.units.convert(blk.unit_model.area, pyo.units.m**2))

def cost_by_flow_volume(blk, flow_cost, flow_to_cost):
    """
    Generic function for costing by flow volume.

    Args:
        flow_cost - The cost of the pump in [currency]/([volume]/[time])
        flow_to_cost - The flow costed in [volume]/[time]
    """
    _make_capital_cost_var(blk)
    blk.flow_cost = pyo.Expression(expr=flow_cost)
    blk.capital_cost_constraint = pyo.Constraint(expr = \
            blk.capital_cost == blk.flow_cost * flow_to_cost)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a landfill unit
operation.
"""

from pyomo.environ import Constraint, units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_pt, constant_intensity, ZeroOrderBaseData

# Some more inforation about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("LandfillZO")
class LandfillZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a landfill unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "landfill"

        build_pt(self)
        constant_intensity(self)

        self.capacity_basis = Var(self.flowsheet().time,
                                units=pyunits.kg/pyunits.hr,
                                doc="capacity basis for capital cost")

        self.total_mass = Var(self.flowsheet().time,
                                units=pyunits.kg/pyunits.hr,
                                doc="total mass flow rate")

        self._fixed_perf_vars.append(self.capacity_basis)

        @self.Constraint(self.flowsheet().time,
                    doc='Total mass constraint')
        def total_mass_constraint(b, t):
            return b.total_mass[t] == sum(pyunits.convert(b.inlet.flow_mass_comp[t, m], to_units=pyunits.kg/pyunits.hr)
                                          for m in b.config.property_package.component_list)

        self._perf_var_dict["Capacity Basis (kg/hr)"] = self.capacity_basis
        self._perf_var_dict["Total Mass (kg/hr)"] = self.total_mass



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import pytest
import watertap.property_models.water_prop_pack as props
from idaes.generic_models.properties.tests.test_harness import \
    PropertyTestHarness as PropertyTestHarness_idaes
from watertap.property_models.tests.property_test_harness import \
    (PropertyTestHarness, PropertyRegressionTest, PropertyCalculateStateTest)

# -----------------------------------------------------------------------------
class TestSeawaterProperty_idaes(PropertyTestHarness_idaes):
    def configure(self):
        self.prop_pack = props.WaterParameterBlock
        self.param_args = {}
        self.prop_args = {}
        self.has_density_terms = False


class TestSeawaterProperty(PropertyTestHarness):
    def configure(self):
        self.prop_pack = props.WaterParameterBlock
        self.param_args = {}
        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1}
        self.stateblock_statistics = {'number_variables': 20,
                                      'number_total_constraints': 16,
                                      'number_unused_variables': 0,
                                      'default_degrees_of_freedom': 4}  # 4 state vars
        self.default_solution = {('dens_mass_phase', 'Liq'): 996.9,
                                 ('dens_mass_phase', 'Vap'): 0.7363,
                                 ('flow_vol_phase', 'Liq'): 5.016e-4,
                                 ('flow_vol_phase', 'Vap'): 0.6790,
                                 ('flow_mol_phase_comp', ('Liq', 'H2O')): 27.75,
                                 ('flow_mol_phase_comp', ('Vap', 'H2O')): 27.75,
                                 ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.5,
                                 ('mole_frac_phase_comp', ('Vap', 'H2O')): 0.5,
                                 ('enth_mass_phase', 'Liq'): 1.049e5,
                                 ('enth_mass_phase', 'Vap'): 2.547e6,
                                 ('dh_vap', None): 2.442e6,
                                 ('cp_phase', 'Liq'): 4.187e3,
                                 ('cp_phase', 'Vap'): 1.865e3,
                                 }


@pytest.mark.component
class TestSeawaterPropertySolution_1(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.WaterParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e8}
        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e-8,
                           ('temperature', None): 273.15 + 50,
                           ('pressure', None): 2e5}
        self.regression_solution = {('dens_mass_phase', 'Liq'): 988.05,
                                    ('dens_mass_phase', 'Vap'): 1.341,
                                    ('flow_vol_phase', 'Liq'): 1.012e-3,
                                    ('flow_vol_phase', 'Vap'): 7.457e-9,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 55.51,
                                    ('flow_mol_phase_comp', ('Vap', 'H2O')): 5.551e-7,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 1,
                                    ('mole_frac_phase_comp', ('Vap', 'H2O')): 1e-8,
                                    ('enth_mass_phase', 'Liq'): 2.094e5,
                                    ('enth_mass_phase', 'Vap'): 2.591e6,
                                    ('dh_vap', None): 2.382e6,
                                    ('cp_phase', 'Liq'): 4.181e3,
                                    ('cp_phase', 'Vap'): 1.871e3,
                                    }


@pytest.mark.component
class TestSeawaterPropertySolution_2(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.WaterParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e8,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1}
        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e-8,
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 1,
                           ('temperature', None): 273.15 + 100,
                           ('pressure', None): 1e5}
        self.regression_solution = {('dens_mass_phase', 'Liq'): 958.3,
                                    ('dens_mass_phase', 'Vap'): 0.5807,
                                    ('flow_vol_phase', 'Liq'): 1.044e-11,
                                    ('flow_vol_phase', 'Vap'): 1.722,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 5.551e-7,
                                    ('flow_mol_phase_comp', ('Vap', 'H2O')): 55.51,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 1e-8,
                                    ('mole_frac_phase_comp', ('Vap', 'H2O')): 1,
                                    ('enth_mass_phase', 'Liq'): 4.190e5,
                                    ('enth_mass_phase', 'Vap'): 2.676e6,
                                    ('dh_vap', None): 2.257e6,
                                    ('cp_phase', 'Liq'): 4.215e3,
                                    ('cp_phase', 'Vap'): 1.890e3,
                                    }



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Feed block for zero-order flowsheets with methods for getting concentration
data from database
"""

from pyomo.environ import (
    check_optimal_termination, Constraint, units as pyunits, value, Var)

from idaes.generic_models.unit_models.feed import FeedData
from idaes.core import declare_process_block_class
import idaes.logger as idaeslog
from idaes.core.util import get_solver
from idaes.core.util.exceptions import InitializationError

# Some more inforation about this module
__author__ = "Andrew Lee"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("FeedZO")
class FeedZOData(FeedData):
    """
    Zero-Order feed block.
    """

    CONFIG = FeedData.CONFIG()

    def build(self):
        super().build()

        units = self.config.property_package.get_metadata().get_derived_units
        comp_list = self.config.property_package.solute_set

        self.flow_vol = Var(self.flowsheet().time,
                            initialize=1,
                            units=units("volume")/units("time"),
                            doc="Volumetric flowrate in feed")

        self.conc_mass_comp = Var(self.flowsheet().time,
                                  comp_list,
                                  initialize=1,
                                  units=units("density_mass"),
                                  doc="Component mass concentrations")

        def rule_Q(blk, t):
            return (self.flow_vol[t] * self.properties[t].dens_mass ==
                    sum(self.properties[t].flow_mass_comp[j]
                        for j in self.properties[t].component_list))
        self.flow_vol_constraint = Constraint(self.flowsheet().time,
                                              rule=rule_Q)

        def rule_C(blk, t, j):
            return (self.conc_mass_comp[t, j] *
                    sum(self.properties[t].flow_mass_comp[k]
                        for k in self.properties[t].component_list) ==
                    self.properties[t].flow_mass_comp[j] *
                    self.properties[t].dens_mass)
        self.conc_mass_constraint = Constraint(self.flowsheet().time,
                                               comp_list,
                                               rule=rule_C)

    def load_feed_data_from_database(self, overwrite=False):
        """
        Method to load initial flowrate and concentrations from database.

        Args:
            overwrite - (default = False), indicates whether fixed values
                        should be overwritten by values from database or not.

        Returns:
            None

        Raises:
            KeyError if flowrate or concentration values not defined in data
        """
        # Get database and water source from property package
        db = self.config.property_package.config.database
        water_source = self.config.property_package.config.water_source

        # Get feed data from database
        data = db.get_source_data(water_source)

        for t in self.flowsheet().time:
            if overwrite or not self.flow_vol[t].fixed:
                try:
                    val = data["default_flow"]["value"]
                    units = getattr(pyunits, data["default_flow"]["units"])
                    self.flow_vol[t].fix(val*units)
                except KeyError:
                    _log.info(
                        f"{self.name} no default flowrate was defined "
                        f"in database water source. Value was not fixed.")

        for (t, j), v in self.conc_mass_comp.items():
            if overwrite or not v.fixed:
                try:
                    val = data["solutes"][j]["value"]
                    units = getattr(pyunits, data["solutes"][j]["units"])
                    v.fix(val*units)
                except KeyError:
                    _log.info(f"{self.name} component {j} was not defined in "
                              f"database water source. Value was not fixed.")

        # Set initial values for mass flows in properties based on these
        # Assuming density of 1000
        for t in self.flowsheet().time:
            for j in self.properties[t].params.solute_set:
                self.properties[t].flow_mass_comp[j].set_value(
                    value(self.flow_vol[t] * self.conc_mass_comp[t, j]))
            self.properties[t].flow_mass_comp["H2O"].set_value(
                value(self.flow_vol[t]*1000))

    def initialize(blk, state_args=None, outlvl=idaeslog.NOTSET,
                   solver=None, optarg=None):
        '''
        This method calls the initialization method of the Feed state block.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) to provide an initial state for
                           initialization (see documentation of the specific
                           property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        '''
        # ---------------------------------------------------------------------
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        if optarg is None:
            optarg = {}

        opt = get_solver(solver, optarg)

        if state_args is None:
            state_args = {}

        # Initialize state block
        blk.properties.initialize(outlvl=outlvl,
                                  optarg=optarg,
                                  solver=solver,
                                  state_args=state_args)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info("Initialization complete: {}.".format(
                idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{blk.name} failed to initialize successfully. Please check "
                f"the output logs for more information.")



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pyomo.environ as pyo
from pyomo.core.base.block import _BlockData
from pyomo.core.kernel.block import IBlock
from pyomo.solvers.plugins.solvers.IPOPT import IPOPT

import idaes.core.util.scaling as iscale
from idaes.core.util.scaling import (
        get_scaling_factor, set_scaling_factor, unset_scaling_factor)
from idaes.logger import getLogger

_log = getLogger("watertap.core")


@pyo.SolverFactory.register("ipopt-watertap",
        doc="The Ipopt NLP solver, with user-based variable and automatic Jacobian constraint scaling")
class IpoptWaterTAP(IPOPT):

    def __init__(self, **kwds):
        kwds["name"] = "ipopt-watertap"
        self._cleanup_needed = False
        super().__init__(**kwds)

    def _presolve(self, *args, **kwds):
        if len(args) > 1 or len(args) == 0:
            raise TypeError(f"IpoptWaterTAP.solve takes 1 positional argument but {len(args)} were given")
        if not isinstance(args[0], (_BlockData, IBlock)):
            raise TypeError("IpoptWaterTAP.solve takes 1 positional argument: a Pyomo ConcreteModel or Block")

        self._tee = kwds.get("tee", False)

        # Set the default watertap options
        if "tol" not in self.options:
            self.options["tol"] = 1e-08
        if "constr_viol_tol" not in self.options:
            self.options["constr_viol_tol"] = 1e-08

        if not self._is_user_scaling():
            self._cleanup_needed = False
            return super()._presolve(*args, **kwds)

        if self._tee:
            print("ipopt-watertap: Ipopt with user variable scaling and IDAES jacobian constraint scaling")

        bound_relax_factor = self._get_option("bound_relax_factor", 1e-10)
        if bound_relax_factor < 0.:
            raise ValueError(f"Option bound_relax_factor must be non-negative; bound_relax_factor={bound_relax_factor}")

        # we are doing this ourselves, don't want Ipopt to also do it
        # also effectively turns "honor_original_bounds" off
        self.options["bound_relax_factor"] = 0.0

        # raise an error if "honor_original_bounds" is set to "yes" (for now)
        if self.options.get("honor_original_bounds", "no") == "yes":
            raise ValueError(f"""Option honor_original_bounds must be set to "no" -- ipopt-watertap does not presently implement this option""")

        # These options are typically available with gradient-scaling, and they
        # have corresponding options in the IDAES constraint_autoscale_large_jac
        # function. Here we use their Ipopt names and default values, see
        # https://coin-or.github.io/Ipopt/OPTIONS.html#OPT_NLP_Scaling
        max_grad = self._get_option("nlp_scaling_max_gradient", 100)
        min_scale = self._get_option("nlp_scaling_min_value", 1e-08)

        # These options are custom for the IDAES constraint_autoscale_large_jac
        # function. We expose them as solver options as this has become part
        # of the solve process.
        ignore_variable_scaling = self._get_option("ignore_variable_scaling", False)
        ignore_constraint_scaling = self._get_option("ignore_constraint_scaling", False)

        self._model = args[0]
        self._cache_scaling_factors()
        self._cache_and_set_relaxed_bounds(bound_relax_factor)
        self._cleanup_needed = True

        # NOTE: This function sets the scaling factors on the
        #       constraints. Hence we cache the constraint scaling
        #       factors and reset them to their original values
        #       so that repeated calls to solve change the scaling
        #       each time based on the initial values, just like in Ipopt.
        try:
            iscale.constraint_autoscale_large_jac(self._model,
                    ignore_constraint_scaling=ignore_constraint_scaling,
                    ignore_variable_scaling=ignore_variable_scaling,
                    max_grad=max_grad,
                    min_scale=min_scale)
        except Exception as err:
            if str(err) == "Error in AMPL evaluation":
                print("ipopt-watertap: Issue in AMPL function evaluation; Jacobian constraint scaling not applied.")
                halt_on_ampl_error = self.options.get("halt_on_ampl_error", "yes")
                if halt_on_ampl_error == "no" :
                    print("ipopt-watertap: halt_on_ampl_error=no, so continuing with optimization.")
                else:
                    self._cleanup()
                    raise RuntimeError("Error in AMPL evaluation.\n"
                            "Run ipopt with halt_on_ampl_error=yes and symbolic_solver_labels=True to see the affected function.")
            else:
                print("Error in constraint_autoscale_large_jac")
                self._cleanup()
                raise

        try:
            # this creates the NL file, among other things
            return super()._presolve(*args, **kwds)
        except:
            self._cleanup()
            raise

    def _cleanup(self):
        if self._cleanup_needed:
            self._reset_scaling_factors()
            self._reset_bounds()
            # remove our reference to the model
            del self._model

    def _postsolve(self):
        self._cleanup()
        return super()._postsolve()

    def _cache_scaling_factors(self):
        self._scaling_cache = [ (c, get_scaling_factor(c)) for c in
                self._model.component_data_objects(
                    pyo.Constraint, active=True, descend_into=True)]

    def _reset_scaling_factors(self):
        for c, s in self._scaling_cache:
            if s is None:
                unset_scaling_factor(c)
            else:
                set_scaling_factor(c, s)
        del self._scaling_cache

    def _cache_and_set_relaxed_bounds(self, bound_relax_factor):
        self._bound_cache = pyo.ComponentMap()
        val = pyo.value
        for v in self._model.component_data_objects(pyo.Var, active=True, descend_into=True):
            # we could hit a variable more
            # than once because of References
            if v in self._bound_cache:
                continue
            if v.lb is None and v.ub is None:
                continue
            self._bound_cache[v] = (v.lb, v.ub)
            sf = get_scaling_factor(v, default=1)
            if v.lb is not None:
                v.lb = val((v.lb*sf - bound_relax_factor*max(1, abs(val(v.lb*sf))))/sf)
            if v.ub is not None:
                v.ub = val((v.ub*sf + bound_relax_factor*max(1, abs(val(v.ub*sf))))/sf)

    def _reset_bounds(self):
        for v, (lb, ub) in self._bound_cache.items():
            v.lb = lb
            v.ub = ub
        del self._bound_cache

    def _get_option(self, option_name, default_value):
        # NOTE: options get reset to their original value at the end of the
        #       OptSolver.solve. The options in _presolve (where this is called)
        #       are already copies of the original, so it is safe to pop them so
        #       they don't get sent to Ipopt.
        option_value = self.options.pop(option_name, None)
        if option_value is None:
            option_value = default_value
        else:
            if self._tee:
                print(f"ipopt-watertap: {option_name}={option_value}")
        return option_value

    def _is_user_scaling(self):
        if "nlp_scaling_method" not in self.options:
            self.options["nlp_scaling_method"] = "user-scaling"
        if self.options["nlp_scaling_method"] != "user-scaling":
            if self._tee:
                print("The ipopt-watertap solver is designed to be run with user-scaling. "
                        f"Ipopt with nlp_scaling_method={self.options['nlp_scaling_method']} will be used instead")
            return False
        return True


## reconfigure IDAES to use the ipopt-watertap solver
import idaes
_default_solver_config_value = idaes.cfg.get("default_solver")
_idaes_default_solver = _default_solver_config_value._default

_default_solver_config_value.set_default_value("ipopt-watertap")
if not _default_solver_config_value._userSet:
    _default_solver_config_value.reset()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import pytest
from io import StringIO

from pyomo.environ import (ConcreteModel,
                           assert_optimal_termination)
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.unit_models.mvc.components import Compressor
import watertap.property_models.water_prop_pack as props

solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.component
def test_compressor():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={'dynamic': False})
    m.fs.properties = props.WaterParameterBlock()
    m.fs.compressor = Compressor(default={"property_package": m.fs.properties})
    # m.fs.compressor.control_volume.display()

    # scaling
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Vap', 'H2O'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
    iscale.set_scaling_factor(m.fs.compressor.control_volume.work, 1e-6)
    iscale.calculate_scaling_factors(m)

    # state variables
    m.fs.compressor.inlet.flow_mass_phase_comp[0, 'Vap', 'H2O'].fix(1)
    m.fs.compressor.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(1e-8)
    m.fs.compressor.inlet.temperature[0].fix(350)  # K
    m.fs.compressor.inlet.pressure[0].fix(0.5e5)  # Pa

    # specifications
    m.fs.compressor.pressure_ratio.fix(2)
    m.fs.compressor.efficiency.fix(0.8)

    # check build
    assert_units_consistent(m)
    assert (degrees_of_freedom(m) == 0)

    # initialize
    m.fs.compressor.initialize()

    # solve
    solver = get_solver()
    results = solver.solve(m, tee=False)
    assert_optimal_termination(results)

    report_io = StringIO()
    m.fs.compressor.report(ostream=report_io)
    output = \
        """
====================================================================================
Unit : fs.compressor                                                       Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key            : Value      : Fixed : Bounds
        Efficiency :    0.80000 :  True : (1e-08, 1)
    Pressure ratio :     2.0000 :  True : (1, 10)
              Work : 1.1534e+05 : False : (None, None)

------------------------------------------------------------------------------------
    Stream Table
                                           Inlet     Outlet  
    flow_mass_phase_comp ('Liq', 'H2O') 1.0000e-08 1.0000e-08
    flow_mass_phase_comp ('Vap', 'H2O')     1.0000     1.0000
    temperature                             350.00     429.57
    pressure                                50000. 1.0000e+05
====================================================================================
"""
    assert output == report_io.getvalue()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pytest
import os
import numpy as np
import pyomo.environ as pyo
import yaml

from pyomo.environ import value

from watertap.tools.parameter_sweep_input_parser import (_yaml_to_dict,
                                                         get_sweep_params_from_yaml,
                                                         set_defaults_from_yaml)

# Imports for conditional fails
from idaes.config import bin_directory as idaes_bin_directory

# -----------------------------------------------------------------------------

class TestInputParser():
    @pytest.fixture(scope="class")
    def model(self):
        m = pyo.ConcreteModel()
        m.fs = pyo.Block()
        m.fs.x = pyo.Var(initialize=-0.5)
        m.fs.a = pyo.Param(initialize=1.0, mutable=True)
        m.fs.b = pyo.Param(initialize=2.0, mutable=False)
        m.fs.sum = pyo.Expression(expr = m.fs.a + m.fs.x)
        m.fs.lim = pyo.Constraint(expr = m.fs.sum >= 0.0)
        m.fs.obj = pyo.Objective(expr = m.fs.x, sense=pyo.minimize)

        return m


    @pytest.fixture(scope="class")
    def get_simple_yaml_file(self):
        filename = 'temp.yaml'
        dict_to_write = {'a_val': 1.1, 'b_val': 2.2, 'c_val': 3.3, 'd_val': 4.4}

        with open(filename, 'w') as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write


    @pytest.fixture(scope="class")
    def get_param_yaml_file(self):
        filename = 'temp.yaml'
        dict_to_write = {'a_val': {'type':'LinearSample',         'param':'fs.a', 'lower_limit':10, 'upper_limit':20, 'num_samples':5},
                         'b_val': {'type':'UniformSample',        'param':'fs.a', 'lower_limit':10, 'upper_limit':20},
                         'c_val': {'type':'NormalSample',         'param':'fs.a', 'mean':10, 'std':2},
                         'd_val': {'type':'LatinHypercubeSample', 'param':'fs.a', 'lower_limit':10, 'upper_limit':20},}

        with open(filename, 'w') as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write


    @pytest.fixture(scope="class")
    def get_param_yaml_file_error(self):
        filename = 'temp.yaml'
        dict_to_write = {'a_val': {'param':'nonexistent_variable'}}

        with open(filename, 'w') as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write


    @pytest.fixture(scope="class")
    def get_default_yaml_file(self, model):
        filename = 'temp.yaml'

        m = model

        dict_to_write = {}

        for pyo_obj in m.component_data_objects((pyo.Var, pyo.Param), active=True):
            for k in range(10):
                dict_to_write[pyo_obj.name] = np.random.rand()

        with open(filename, 'w') as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write


    @pytest.fixture(scope="class")
    def get_default_yaml_file_error(self, model):
        filename = 'temp.yaml'

        dict_to_write = {'nonexistent_variable': 1.0}

        with open(filename, 'w') as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write


    @pytest.mark.unit
    def test_yaml_to_dict(self, get_simple_yaml_file):
        filename, truth_values = get_simple_yaml_file

        values = _yaml_to_dict(filename)

        for key, truth_value in truth_values.items():
            assert key in values
            assert values[key] == truth_value


    @pytest.mark.unit
    def test_yaml_to_dict_error(self):
        with pytest.raises(Exception):
            values = _yaml_to_dict('nonexistent_file.yaml')


    @pytest.mark.unit
    def test_get_sweep_params_from_yaml(self, model, get_param_yaml_file):
        m = model
        filename, truth_values = get_param_yaml_file

        values = get_sweep_params_from_yaml(m, filename)

        for key, truth_value in truth_values.items():
            assert key in values


    @pytest.mark.unit
    def test_get_sweep_params_from_yaml_error(self, model, get_param_yaml_file_error):
        m = model
        filename, truth_values = get_param_yaml_file_error

        with pytest.raises(ValueError):
            get_sweep_params_from_yaml(m, filename)


    @pytest.mark.unit
    def test_set_defaults_from_yaml(self, model, get_default_yaml_file):
        m = model
        filename, truth_values = get_default_yaml_file

        set_defaults_from_yaml(m, filename)

        for key, truth_value in truth_values.items():
            component = m.find_component(key)
            if component is m.fs.b: # don't set non-mutable params
                assert value(component) == 2.0
            else:
                assert value(component) == truth_value


    @pytest.mark.unit
    def test_set_defaults_from_yaml_error(self, model, get_default_yaml_file_error):
        m = model
        filename, truth_values = get_default_yaml_file_error

        with pytest.raises(ValueError):
            set_defaults_from_yaml(m, filename)

        os.remove(filename)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a Chlorination unit.
"""

from pyomo.environ import units as pyunits, Var
from pyomo.common.config import ConfigValue, In
from idaes.core import declare_process_block_class
from watertap.core import build_siso, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("ChlorinationZO")
class ChlorinationZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a Chlorination unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "chlorination"

        build_siso(self)
        constant_intensity(self)

        self.initial_chlorine_demand = Var(self.flowsheet().time,
                                           units=pyunits.mg/pyunits.liter,
                                           doc="Initial chlorine demand")

        self.contact_time = Var(self.flowsheet().time,
                                units=pyunits.hour,
                                doc="Chlorine contact time")
        self.concentration_time = Var(
            self.flowsheet().time,
            units=(pyunits.mg*pyunits.minute)/pyunits.liter,
            doc="CT value for chlorination")
        self.chlorine_decay_rate = Var(
            self.flowsheet().time,
            units=pyunits.mg/(pyunits.L*pyunits.hour),
            doc="Chlorine decay rate")

        self.recovery_frac_mass_H2O.fix(1)
        self._fixed_perf_vars.append(self.initial_chlorine_demand)
        self._fixed_perf_vars.append(self.contact_time)
        self._fixed_perf_vars.append(self.concentration_time)
        self._fixed_perf_vars.append(self.chlorine_decay_rate)

        self.chlorine_dose = Var(self.flowsheet().time,
                                 units=pyunits.mg/pyunits.L,
                                 doc="Chlorine dose")

        @self.Constraint(self.flowsheet().time,
                         doc="Chlorine dose constraint")
        def chlorine_dose_constraint(b, t):
            return (b.chlorine_dose[t] ==
                    self.initial_chlorine_demand[t] +
                    self.chlorine_decay_rate[t]*self.contact_time[t] +
                    (self.concentration_time[t] /
                     pyunits.convert(self.contact_time[t],
                                     to_units=pyunits.minute)))

        self._perf_var_dict["Chlorine Dose (mg/L)"] = self.chlorine_dose
        self._perf_var_dict["Initial Chlorine Demand (mg/L)"] = \
            self.initial_chlorine_demand
        self._perf_var_dict["Contact Time (hr)"] = self.contact_time
        self._perf_var_dict["CT Value ((mg*min)/L)"] = self.concentration_time
        self._perf_var_dict["Chlorine Decay Rate (mg/(L*hr))"] = \
            self.chlorine_decay_rate



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order fixed bed model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    Block, check_optimal_termination, ConcreteModel, Constraint, value, Var, Param)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import FixedBedZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestFixedBedZO_w_o_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["bod"]})

        m.fs.unit = FixedBedZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "bod"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)

        assert isinstance(model.fs.unit.acetic_acid_dose, Var)
        assert isinstance(model.fs.unit.acetic_acid_demand, Var)
        assert isinstance(model.fs.unit.acetic_acid_demand_equation,
                          Constraint)

        assert isinstance(model.fs.unit.phosphoric_acid_dose, Var)
        assert isinstance(model.fs.unit.phosphoric_acid_demand, Var)
        assert isinstance(model.fs.unit.acetic_acid_demand_equation,
                          Constraint)

        assert isinstance(model.fs.unit.ferric_chloride_dose, Var)
        assert isinstance(model.fs.unit.ferric_chloride_demand, Var)
        assert isinstance(model.fs.unit.ferric_chloride_demand_equation,
                          Constraint)

        assert isinstance(model.fs.unit.activated_carbon_parameter_a, Var)
        assert isinstance(model.fs.unit.activated_carbon_parameter_b, Var)
        assert isinstance(model.fs.unit.activated_carbon_demand, Var)
        assert isinstance(model.fs.unit.activated_carbon_demand_equation,
                          Constraint)

        assert isinstance(model.fs.unit.sand_parameter_a, Var)
        assert isinstance(model.fs.unit.sand_parameter_b, Var)
        assert isinstance(model.fs.unit.sand_demand, Var)
        assert isinstance(model.fs.unit.sand_demand_equation,
                          Constraint)

        assert isinstance(model.fs.unit.anthracite_parameter_a, Var)
        assert isinstance(model.fs.unit.anthracite_parameter_b, Var)
        assert isinstance(model.fs.unit.anthracite_demand, Var)
        assert isinstance(model.fs.unit.anthracite_demand_equation,
                          Constraint)

        assert isinstance(model.fs.unit.cationic_polymer_parameter_a, Var)
        assert isinstance(model.fs.unit.cationic_polymer_parameter_b, Var)
        assert isinstance(model.fs.unit.cationic_polymer_demand, Var)
        assert isinstance(model.fs.unit.cationic_polymer_demand_equation,
                          Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("fixed_bed")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1
        model.fs.unit.load_parameters_from_database()
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

        assert model.fs.unit.acetic_acid_dose.fixed
        assert model.fs.unit.acetic_acid_dose.value == data[
            "acetic_acid_dose"]["value"]
        assert model.fs.unit.phosphoric_acid_dose.fixed
        assert model.fs.unit.phosphoric_acid_dose.value == data[
            "phosphoric_acid_dose"]["value"]
        assert model.fs.unit.ferric_chloride_dose.fixed
        assert model.fs.unit.ferric_chloride_dose.value == data[
            "ferric_chloride_dose"]["value"]

        assert model.fs.unit.activated_carbon_parameter_a.fixed
        assert model.fs.unit.activated_carbon_parameter_a.value == data[
            "activated_carbon_parameter_a"]["value"]
        assert model.fs.unit.activated_carbon_parameter_b.fixed
        assert model.fs.unit.activated_carbon_parameter_b.value == data[
            "activated_carbon_parameter_b"]["value"]

        assert model.fs.unit.sand_parameter_a.fixed
        assert model.fs.unit.sand_parameter_a.value == data[
            "sand_parameter_a"]["value"]
        assert model.fs.unit.sand_parameter_b.fixed
        assert model.fs.unit.sand_parameter_b.value == data[
            "sand_parameter_b"]["value"]

        assert model.fs.unit.anthracite_parameter_a.fixed
        assert model.fs.unit.anthracite_parameter_a.value == data[
            "anthracite_parameter_a"]["value"]
        assert model.fs.unit.anthracite_parameter_b.fixed
        assert model.fs.unit.anthracite_parameter_b.value == data[
            "anthracite_parameter_b"]["value"]

        assert model.fs.unit.cationic_polymer_parameter_a.fixed
        assert model.fs.unit.cationic_polymer_parameter_a.value == data[
            "cationic_polymer_parameter_a"]["value"]
        assert model.fs.unit.cationic_polymer_parameter_b.fixed
        assert model.fs.unit.cationic_polymer_parameter_b.value == data[
            "cationic_polymer_parameter_b"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(10.0001, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(9.9999e-3, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["bod"]))
        assert (pytest.approx(3082.56, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))
        assert (model.fs.unit.properties_in[0].flow_mass_comp["H2O"].value ==
                model.fs.unit.properties_treated[0].flow_mass_comp["H2O"].value)

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                     : Value    : Fixed : Bounds
         Acetic Acid Demand :   1799.0 : False : (0, None)
    Activated Carbon Demand :   58.448 : False : (0, None)
          Anthracite Demand :   64.048 : False : (0, None)
    Cationic Polymer Demand :   49.057 : False : (0, None)
         Electricity Demand :   3082.6 : False : (0, None)
      Electricity Intensity : 0.085618 :  True : (None, None)
    Ferric Chlorided Demand :   359.80 : False : (0, None)
     Phosphoric Acid Demand :   152.17 : False : (0, None)
                Sand Demand :   64.897 : False : (0, None)
       Solute Removal [bod] :  0.90000 :  True : (0, None)

------------------------------------------------------------------------------------
    Stream Table
                             Inlet    Treated
    Volumetric Flowrate      10.001    10.000
    Mass Concentration H2O   999.90    999.99
    Mass Concentration bod 0.099990 0.0099999
====================================================================================
"""

        assert output in stream.getvalue()


class TestFixedBedZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["bod", "foo"]})

        m.fs.unit = FixedBedZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "bod"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(4)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("fixed_bed")

        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1
        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

        assert model.fs.unit.acetic_acid_dose.fixed
        assert model.fs.unit.acetic_acid_dose.value == data[
            "acetic_acid_dose"]["value"]
        assert model.fs.unit.phosphoric_acid_dose.fixed
        assert model.fs.unit.phosphoric_acid_dose.value == data[
            "phosphoric_acid_dose"]["value"]
        assert model.fs.unit.ferric_chloride_dose.fixed
        assert model.fs.unit.ferric_chloride_dose.value == data[
            "ferric_chloride_dose"]["value"]

        assert model.fs.unit.activated_carbon_parameter_a.fixed
        assert model.fs.unit.activated_carbon_parameter_a.value == data[
            "activated_carbon_parameter_a"]["value"]
        assert model.fs.unit.activated_carbon_parameter_b.fixed
        assert model.fs.unit.activated_carbon_parameter_b.value == data[
            "activated_carbon_parameter_b"]["value"]

        assert model.fs.unit.sand_parameter_a.fixed
        assert model.fs.unit.sand_parameter_a.value == data[
            "sand_parameter_a"]["value"]
        assert model.fs.unit.sand_parameter_b.fixed
        assert model.fs.unit.sand_parameter_b.value == data[
            "sand_parameter_b"]["value"]

        assert model.fs.unit.anthracite_parameter_a.fixed
        assert model.fs.unit.anthracite_parameter_a.value == data[
            "anthracite_parameter_a"]["value"]
        assert model.fs.unit.anthracite_parameter_b.fixed
        assert model.fs.unit.anthracite_parameter_b.value == data[
            "anthracite_parameter_b"]["value"]

        assert model.fs.unit.cationic_polymer_parameter_a.fixed
        assert model.fs.unit.cationic_polymer_parameter_a.value == data[
            "cationic_polymer_parameter_a"]["value"]
        assert model.fs.unit.cationic_polymer_parameter_b.fixed
        assert model.fs.unit.cationic_polymer_parameter_b.value == data[
            "cationic_polymer_parameter_b"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(10.00410, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(9.99590e-3, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["bod"]))
        assert (pytest.approx(0.39984, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(3083.79, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))
        assert (model.fs.unit.properties_in[0].flow_mass_comp["H2O"].value ==
                model.fs.unit.properties_treated[0].flow_mass_comp["H2O"].value)

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                     : Value    : Fixed : Bounds
         Acetic Acid Demand :   1799.7 : False : (0, None)
    Activated Carbon Demand :   58.469 : False : (0, None)
          Anthracite Demand :   64.072 : False : (0, None)
    Cationic Polymer Demand :   49.073 : False : (0, None)
         Electricity Demand :   3083.8 : False : (0, None)
      Electricity Intensity : 0.085618 :  True : (None, None)
    Ferric Chlorided Demand :   359.95 : False : (0, None)
     Phosphoric Acid Demand :   152.23 : False : (0, None)
                Sand Demand :   64.921 : False : (0, None)
       Solute Removal [bod] :  0.90000 :  True : (0, None)
       Solute Removal [foo] :   0.0000 :  True : (0, None)

------------------------------------------------------------------------------------
    Stream Table
                             Inlet    Treated
    Volumetric Flowrate      10.005    10.004
    Mass Concentration H2O   999.50    999.59
    Mass Concentration bod 0.099950 0.0099959
    Mass Concentration foo  0.39980   0.39984
====================================================================================
"""
        assert output in stream.getvalue()


db = Database()
params = db._get_technology("fixed_bed")


class TestIXZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["bod"]})

        m.fs.unit = FixedBedZO(default={
            "property_package": m.fs.params,
            "database": db})

        return m

    @pytest.mark.parametrize("subtype", [params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("fixed_bed", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]


db = Database()
params = db._get_technology("fixed_bed")


@pytest.mark.component
@pytest.mark.parametrize("subtype", [k for k in params.keys()])
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = FixedBedZO(default={
        "property_package": m.fs.params,
        "database": m.db,
        "process_subtype": subtype})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.fixed_bed, Block)
    assert isinstance(m.fs.costing.fixed_bed.capital_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.fixed_bed.capital_b_parameter,
                      Var)
    assert isinstance(m.fs.costing.fixed_bed.reference_state, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]
    assert m.fs.unit1.acetic_acid_demand[0] in \
        m.fs.costing._registered_flows["acetic_acid"]
    assert m.fs.unit1.phosphoric_acid_demand[0] in \
        m.fs.costing._registered_flows["phosphoric_acid"]
    assert m.fs.unit1.ferric_chloride_demand[0] in \
        m.fs.costing._registered_flows["ferric_chloride"]
    assert m.fs.unit1.activated_carbon_demand[0] in \
        m.fs.costing._registered_flows["activated_carbon"]
    assert m.fs.unit1.sand_demand[0] in \
        m.fs.costing._registered_flows["sand"]
    assert m.fs.unit1.anthracite_demand[0] in \
        m.fs.costing._registered_flows["anthracite"]
    assert m.fs.unit1.cationic_polymer_demand[0] in \
        m.fs.costing._registered_flows["cationic_polymer"]



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pytest
from pyomo.environ import (ConcreteModel,
                           Constraint,
                           TerminationCondition,
                           SolverStatus,
                           value,
                           Var,
                           units as pyunits,
                           assert_optimal_termination)
from pyomo.network import Port
from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        ControlVolume0DBlock)
from watertap.property_models.ion_DSPMDE_prop_pack import DSPMDEParameterBlock, ActivityCoefficientModel, \
    DensityCalculation, DSPMDEStateBlock
from watertap.unit_models.nanofiltration_DSPMDE_0D import NanofiltrationDSPMDE0D, MassTransferCoefficient
from watertap.core.util.initialization import check_dof

from idaes.core.util import get_solver
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              number_unused_variables,
                                              unused_variables_set)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (calculate_scaling_factors,
                                     unscaled_variables_generator,
                                     unscaled_constraints_generator,
                                     constraints_with_scale_factor_generator,
                                     badly_scaled_var_generator)
# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={'dynamic': False})
    m.fs.properties = DSPMDEParameterBlock(default={
            "solute_list": ["Ca_2+", "SO4_2-", "Na_+", "Cl_-", "Mg_2+"]})
    m.fs.unit = NanofiltrationDSPMDE0D(default={'property_package': m.fs.properties})


    # check unit config arguments
    assert len(m.fs.unit.config) == 8

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == \
           MaterialBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == \
           MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is \
           m.fs.properties
    assert m.fs.unit.config.property_package.config.activity_coefficient_model == ActivityCoefficientModel.ideal
    assert m.fs.unit.config.property_package.config.density_calculation == DensityCalculation.constant
    assert m.fs.unit.config.mass_transfer_coefficient == MassTransferCoefficient.fixed

class TestNanoFiltration():
    @pytest.fixture(scope="class")
    def NF_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={'dynamic': False})
        m.fs.properties = DSPMDEParameterBlock(default={
            "solute_list": [
                "Ca_2+",
                "SO4_2-",
                "Mg_2+",
                "Na_+",
                "Cl_-",
            ],
            "diffusivity_data": {
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "SO4_2-"): 1.06e-9,
                ("Liq", "Mg_2+"): 0.706e-9,
                ("Liq", "Na_+"): 1.33e-9,
                ("Liq", "Cl_-"): 2.03e-9,
            },
            "mw_data": {
                "H2O": 18e-3,
                "Ca_2+": 40e-3,
                "Mg_2+": 24e-3,
                "SO4_2-": 96e-3,
                "Na_+": 23e-3,
                "Cl_-": 35e-3,
            },
            "stokes_radius_data": {
                "Ca_2+": 0.309e-9,
                "Mg_2+": 0.347e-9,
                "SO4_2-": 0.230e-9,
                "Cl_-": 0.121e-9,
                "Na_+": 0.184e-9,
            },
            "charge": {
                "Ca_2+": 2,
                "Mg_2+": 2,
                "SO4_2-": -2,
                "Na_+": 1,
                "Cl_-": -1,
            },
            "activity_coefficient_model": ActivityCoefficientModel.davies,
            "density_calculation": DensityCalculation.constant
        })

        m.fs.unit = NanofiltrationDSPMDE0D(default={"property_package": m.fs.properties,
                                                    "mass_transfer_coefficient": MassTransferCoefficient.spiral_wound})
        b = m.fs.unit
        mass_flow_in = 1 * pyunits.kg / pyunits.s
        feed_mass_frac = {
            'Ca_2+': 382e-6,
            'Mg_2+': 1394e-6,
            'SO4_2-': 2136e-6,
            'Cl_-': 20101.6e-6,
            'Na_+': 11122e-6,
        }

        # Fix mole flow rates of each ion and water
        for ion, x in feed_mass_frac.items():
            mol_comp_flow = x * pyunits.kg / pyunits.kg * mass_flow_in / m.fs.unit.feed_side.properties_in[0].mw_comp[
                ion]
            m.fs.unit.inlet.flow_mol_phase_comp[0, 'Liq', ion].fix(mol_comp_flow)
        H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
        H2O_mol_comp_flow = H2O_mass_frac * pyunits.kg / pyunits.kg * mass_flow_in / \
                            m.fs.unit.feed_side.properties_in[0].mw_comp['H2O']
        m.fs.unit.inlet.flow_mol_phase_comp[0, 'Liq', 'H2O'].fix(H2O_mol_comp_flow)

        # Use assert electroneutrality method from property model to ensure the ion concentrations provided
        # obey electroneutrality condition
        m.fs.unit.feed_side.properties_in[0].assert_electroneutrality(defined_state=True,
                                                                      adjust_by_ion='Cl_-',
                                                                      get_property='mass_frac_phase_comp')

        # Fix other inlet state variables
        m.fs.unit.inlet.temperature[0].fix(298.15)
        m.fs.unit.inlet.pressure[0].fix(4e5)

        # Fix the membrane variables that are usually fixed for the DSPM-DE model
        m.fs.unit.radius_pore.fix(0.5e-9)
        m.fs.unit.membrane_thickness_effective.fix(1.33e-6)
        m.fs.unit.membrane_charge_density.fix(-27)
        m.fs.unit.dielectric_constant_pore.fix(41.3)

        # Fix final permeate pressure to be ~atmospheric
        m.fs.unit.mixed_permeate[0].pressure.fix(101325)

        # Fix additional variables for calculating mass transfer coefficient with spiral wound correlation
        m.fs.unit.spacer_porosity.fix(0.85)
        m.fs.unit.spacer_mixing_efficiency.fix()
        m.fs.unit.spacer_mixing_length.fix()
        m.fs.unit.channel_height.fix(5e-4)
        m.fs.unit.recovery_vol_phase[0, 'Liq'].fix(0.5)

        return m

    @pytest.mark.unit
    def test_build(self, NF_frame):
        m = NF_frame

        # test ports and variables
        port_lst = ['inlet', 'retentate', 'permeate']
        port_vars_lst = ['flow_mol_phase_comp', 'pressure', 'temperature']
        for port_str in port_lst:
            assert hasattr(m.fs.unit, port_str)
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)
            for var_str in port_vars_lst:
                assert hasattr(port, var_str)
                var = getattr(port, var_str)
                assert isinstance(var, Var)

        # test unit objects (including parameters, variables, and constraints)
        unit_objs_lst = []
        for obj_str in unit_objs_lst:
            assert hasattr(m.fs.unit, obj_str)


        # test state block objects
        assert isinstance(m.fs.unit.feed_side, ControlVolume0DBlock)
        cv_stateblock_lst = ['properties_in',
                             'properties_out',
                             'properties_interface',
                            ]
        # feed side
        for sb_str in cv_stateblock_lst:
            sb = getattr(m.fs.unit.feed_side, sb_str)
            assert isinstance(sb, DSPMDEStateBlock)
        # test objects added to control volume
        cv_objs_type_dict = {'eq_feed_interface_isothermal': Constraint}
        for (obj_str, obj_type) in cv_objs_type_dict.items():
            obj = getattr(m.fs.unit.feed_side, obj_str)
            assert isinstance(obj, obj_type)
        # permeate side
        assert isinstance(m.fs.unit.permeate_side, DSPMDEStateBlock)
        assert isinstance(m.fs.unit.mixed_permeate, DSPMDEStateBlock)
        # membrane
        assert isinstance(m.fs.unit.pore_entrance, DSPMDEStateBlock)
        assert isinstance(m.fs.unit.pore_exit, DSPMDEStateBlock)

        # test statistics
        assert number_variables(m) == 558
        assert number_total_constraints(m) == 524
        assert number_unused_variables(m) == 11

    @pytest.mark.unit
    def test_dof(self, NF_frame):
        m = NF_frame
        check_dof(m, fail_flag=True)

    @pytest.mark.unit
    def test_calculate_scaling(self, NF_frame):
        m = NF_frame

        m.fs.properties.set_default_scaling('flow_mol_phase_comp', 1e5, index=('Liq', 'Ca_2+'))
        m.fs.properties.set_default_scaling('flow_mol_phase_comp', 1e5, index=('Liq', 'SO4_2-'))

        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit))
        assert len(unscaled_var_list) == 0

        badly_scaled_var_lst = list(badly_scaled_var_generator(m, include_fixed=True))
        assert len(unscaled_var_list) == 0

        # not all constraints have scaling factor so skipping the check for unscaled constraints

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_initialize(self, NF_frame):
        m= NF_frame
        initialization_tester(m)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_solve(self, NF_frame):
        m = NF_frame
        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_conservation(self, NF_frame):
        m = NF_frame
        b = m.fs.unit
        comp_lst = m.fs.properties.solute_set

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp['Liq', j] for j in comp_lst)
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp['Liq', j] for j in comp_lst)
        flow_mass_permeate = sum(
            b.mixed_permeate[0].flow_mass_phase_comp['Liq', j] for j in comp_lst)

        assert (abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate
                          )) <= 1e-6)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_solution(self, NF_frame):
        m = NF_frame

        mole_flux_dict = {'Na_+': 0.0036645,
                          'Cl_-': 0.004354,
                          'Ca_2+': 7.3313e-5,
                          'SO4_2-': 1.6845e-4,
                          'Mg_2+': 0.000440,
                          'H2O': 0.41437
                          }
        for j, val in mole_flux_dict.items():
            assert (pytest.approx(val, rel=5e-2) ==
                    value(m.fs.unit.flux_mol_phase_comp_avg[0, 'Liq', j]))

        #TODO: subsequently focus on the segment below during the validation and refinement phase
        intrinsic_rejection_dict = {'Na_+': 0.017432,
                                    'Cl_-': 0.015837,
                                    'Ca_2+': -0.023145,
                                    'SO4_2-': 0.015924,
                                    'Mg_2+': 0.01546,
                                    }
        for j, val in intrinsic_rejection_dict.items():
            assert (pytest.approx(val, rel=5e-2) ==
                    value(m.fs.unit.rejection_intrinsic_phase_comp[0, 'Liq', j]))

    @pytest.mark.unit
    def test_report(self, NF_frame):
        NF_frame.fs.unit.report()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order backwash solids handling model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    Block, check_optimal_termination, ConcreteModel, Constraint, value, Var, Param)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import BackwashSolidsHandlingZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestBackwashSolidsHandling_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["nitrate", "tss", "toc", "foo"]})

        m.fs.unit = BackwashSolidsHandlingZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "nitrate"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.lift_height, Param)
        assert isinstance(model.fs.unit.eta_pump, Param)
        assert isinstance(model.fs.unit.eta_motor, Param)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("backwash_solids_handling")
        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
               data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(9.5011, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.0052625, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["nitrate"]))
        assert (pytest.approx(0.0052625, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.0052625, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["toc"]))
        assert (pytest.approx(0.10525, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0.50285, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(1.88923, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["nitrate"]))
        assert (pytest.approx(1.88923, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(1.88923, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["toc"]))
        assert (pytest.approx(1.9887e-08, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]))
        assert (pytest.approx(3686.342, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))


    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                      : Value   : Fixed : Bounds
          Electricity Demand :  3686.3 : False : (0, None)
        Solute Removal [foo] :  0.0000 :  True : (0, None)
    Solute Removal [nitrate] : 0.95000 :  True : (0, None)
        Solute Removal [toc] : 0.95000 :  True : (0, None)
        Solute Removal [tss] : 0.95000 :  True : (0, None)
              Water Recovery : 0.95000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                 Inlet    Treated  Byproduct
    Volumetric Flowrate          10.004    9.5011    0.50285
    Mass Concentration H2O       999.60    999.88     994.33
    Mass Concentration nitrate 0.099960 0.0052625     1.8892
    Mass Concentration tss     0.099960 0.0052625     1.8892
    Mass Concentration toc     0.099960 0.0052625     1.8892
    Mass Concentration foo     0.099960   0.10525 1.9887e-08
====================================================================================
"""
        assert output in stream.getvalue()


db = Database()
params = db._get_technology("backwash_solids_handling")


class TestIXZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["nitrate"]})

        m.fs.unit = BackwashSolidsHandlingZO(default={
            "property_package": m.fs.params,
            "database": db})

        return m

    @pytest.mark.parametrize("subtype", [params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("backwash_solids_handling", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = BackwashSolidsHandlingZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.backwash_solids_handling, Block)
    assert isinstance(
        m.fs.costing.backwash_solids_handling.capital_a_parameter, Var)
    assert isinstance(
        m.fs.costing.backwash_solids_handling.capital_b_parameter, Var)
    assert isinstance(m.fs.costing.backwash_solids_handling.reference_state,
                      Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]
